import re
import logging

from sixpack import sixpack
from django.conf import settings
from requests.exceptions import RequestException

RE_FIRST_CAP = re.compile('(.)([A-Z][a-z]+)')
RE_ALL_CAP = re.compile('([a-z0-9])([A-Z])')
RE_TEST_NAME = re.compile('_test$')

logger = logging.getLogger('djsixpack')


class AlternativesAttributeSetterMeta(type):
    def __new__(mcs, name, bases, dct):
        alts = dct.get('alternatives', ()) or ()
        dct.update(dict((alt, alt) for alt in alts))
        return super(AlternativesAttributeSetterMeta, mcs).__new__(mcs, name, bases, dct)


class SixpackTest(object):
    __metaclass__ = AlternativesAttributeSetterMeta

    unique_attr = 'pk'
    host = None
    timeout = None
    control = None
    alternatives = None

    def __init__(self, instance):
        self._instance = instance
        self.host = self.host or getattr(settings, 'SIXPACK_HOST', sixpack.SIXPACK_HOST)
        self.timeout = self.timeout or getattr(settings, 'SIXPACK_TIMEOUT', sixpack.SIXPACK_TIMEOUT)

    @property
    def client_id(self):
        return self.get_client_id(self._instance)

    def _get_session(self, user_agent=None, ip_address=None):
        session_kwargs = {
            'options': {'host': self.host, 'timeout': self.timeout},
            'params': {'ip_address': ip_address, 'user_agent': user_agent},
            'client_id': self.client_id,
        }

        return sixpack.Session(**session_kwargs)

    def _get_experiment_name(self):
        # Converts CamelCase to snake_case and drops the trailing "_test"
        s1 = RE_FIRST_CAP.sub(r'\1_\2', self.__class__.__name__)
        name = RE_ALL_CAP.sub(r'\1_\2', s1).lower()
        return RE_TEST_NAME.sub('', name)

    def get_client_id(self, instance):
        if not self.unique_attr:
            raise ValueError('Need a unique_attr to compute the client ID')

        client_id = getattr(instance, self.unique_attr, None)
        if not client_id:
            raise ValueError('Unique_attr {0} does not yield a usable identifier'.format(self.unique_attr))

        return client_id

    def participate(self, force=None, user_agent=None, ip_address=None):
        if not self.host:
            try:
                if force in self.alternatives:
                    return force
                return self.alternatives[0]
            except TypeError:
                raise ValueError('No alternatives defined')

        session = self._get_session(user_agent, ip_address)
        experiment_name = self._get_experiment_name()

        try:
            resp = session.participate(experiment_name, self.alternatives, force)
        except RequestException:
            logger.exception("Error while trying to .participate")
            if force in self.alternatives:
                return force
            return self.alternatives[0]
        else:
            return resp['alternative']['name']

    def convert(self, kpi=None):
        if not self.host:
            return True

        session = self._get_session()
        experiment_name = self._get_experiment_name()
        try:
            resp = session.convert(experiment_name)
        except RequestException as e:
            logger.exception("Error while trying to .convert: {err}".format(err=e))
            return False
        else:
            return resp['status'] == 'ok'



from setuptools import setup, find_packages

install_reqs = ["Django>=1.4", "sixpack-client==1.0.0"]

setup(
    name='django-sixpack',
    version=__import__('djsixpack').__version__,
    author="Dan Langer",
    author_email='opensource@waveapps.com',
    include_package_data=True,
    zip_safe=True,
    packages=find_packages(),
    url='https://github.com/waveaccounting/django-sixpack',
    license='LICENSE',
    description='A django-friendly wrapper for sixpack-py',
    long_description=open('README.rst').read(),
    install_requires=install_reqs,
    tests_require=install_reqs + ["mock==1.0.1"],
    test_suite='runtests.runtests',
    classifiers=(
        'Development Status :: 5 - Production/Stable',
        'Framework :: Django',
        'Natural Language :: English',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ),
)



from mock import Mock, patch, call

from django.test import TestCase

from djsixpack.djsixpack import SixpackTest


class ClientIdTest(TestCase):

    def test_client_id_property_calls_method(self):
        t = SixpackTest(Mock())
        t.get_client_id = Mock()

        t.client_id
        self.assertTrue(t.get_client_id.called)

    def test_client_id_no_attr_set(self):
        t = SixpackTest(Mock())
        t.unique_attr = None

        with self.assertRaises(ValueError):
            t.client_id

    def test_client_id_no_usable_identifier(self):
        t = SixpackTest(Mock(pk=None))

        with self.assertRaises(ValueError):
            t.client_id

    def test_client_id_uses_set_attr(self):
        t = SixpackTest(Mock(pk=5))

        self.assertEqual(t.client_id, 5)


class ParticipateTest(TestCase):

    def test_participate_errors_when_no_alternatives(self):
        mock_user = Mock(pk=10)

        class NoAltsTest(SixpackTest):
            alternatives = ()

        expt = NoAltsTest(mock_user)
        self.assertRaises(ValueError, expt.participate)

    def test_participate_returns_default_when_no_host_set(self):
        mock_user = Mock(pk=10)

        class DefaultTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        with patch('djsixpack.djsixpack.sixpack') as sp_mock:
            with self.settings(SIXPACK_HOST=None):
                expt = DefaultTest(mock_user)
                alternative = expt.participate()

        self.assertEqual(alternative, 'FIRST')
        self.assertFalse(sp_mock.participate.called)

    def test_participate_returns_force_when_no_host_set_and_force_in_alternatives(self):
        mock_user = Mock(pk=10)

        class DefaultTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        with patch('djsixpack.djsixpack.sixpack') as sp_mock:
            with self.settings(SIXPACK_HOST=None):
                expt = DefaultTest(mock_user)
                alternative = expt.participate(force='SECOND')

        self.assertEqual(alternative, 'SECOND')
        self.assertFalse(sp_mock.participate.called)

    def test_participate_returns_default_when_no_host_set_and_force_not_in_alternatives(self):
        mock_user = Mock(pk=10)

        class DefaultTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        with patch('djsixpack.djsixpack.sixpack') as sp_mock:
            with self.settings(SIXPACK_HOST=None):
                expt = DefaultTest(mock_user)
                alternative = expt.participate(force='THIRD')

        self.assertEqual(alternative, 'FIRST')
        self.assertFalse(sp_mock.participate.called)

    def test_participate_calls_library(self):
        mock_user = Mock(pk=10)

        class DefaultTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        with patch('djsixpack.djsixpack.sixpack') as sp_mock:
            expt = DefaultTest(mock_user)
            expt.participate()

        self.assertEqual(
            sp_mock.Session.call_args_list,
            [call(
                params={'ip_address': None, 'user_agent': None},
                options={'host': sp_mock.SIXPACK_HOST, 'timeout': sp_mock.SIXPACK_TIMEOUT},
                client_id=10
            )]
        )
        self.assertEqual(
            sp_mock.Session.return_value.participate.call_args_list,
            [call('default', ('FIRST', 'SECOND'), None)]
        )

        class ConvertTest(TestCase):

            def test_convert_doesnt_call_library_when_no_host(self):
                Mock(pk=10)

        class DefaultTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        with patch('djsixpack.djsixpack.sixpack') as sp_mock:
            with self.settings(SIXPACK_HOST=None):
                expt = DefaultTest(mock_user)
                expt.convert()

        self.assertFalse(sp_mock.Session().convert.called)

    def test_convert_calls_library(self):
        mock_user = Mock(pk=10)

        class DefaultTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        with patch('djsixpack.djsixpack.sixpack') as sp_mock:
            expt = DefaultTest(mock_user)
            expt.convert(kpi='cats')

        self.assertEqual(
            sp_mock.Session().convert.call_args_list,
            [call('default')]
        )


class ExperimentNameTest(TestCase):

    def test_to_snake(self):
        class CompoundNameTest(SixpackTest):
            pass

        expt = CompoundNameTest(Mock)
        self.assertEqual(expt._get_experiment_name(), 'compound_name')

    def test_name_drops_test(self):
        class CompoundNameTest(SixpackTest):
            pass

        expt = CompoundNameTest(Mock)
        self.assertNotIn('_test', expt._get_experiment_name())


class AlternativesSetAsAttributesTest(TestCase):

    def test_alternatives_are_set_as_attributes(self):
        class AltTest(SixpackTest):
            alternatives = ('FIRST', 'SECOND')

        self.assertTrue(hasattr(AltTest, 'FIRST'))
        self.assertTrue(hasattr(AltTest, 'SECOND'))
